package geoiphandler

import (
	"context"
	"fmt"
	"net"
	"path/filepath"
	"strings"
	"sync"
	"time"

	"go-micro.dev/v4/errors"
	"go-micro.dev/v4/server"
	"jochum.dev/jo-micro/auth2"
	"jochum.dev/jo-micro/auth2/plugins/verifier/endpointroles"
	"jochum.dev/jo-micro/geoip/internal/maxmind"
	"jochum.dev/jo-micro/geoip/proto/geoippb"
	"jochum.dev/jo-micro/geoip/utils"
	"jochum.dev/jo-micro/logruscomponent"
	"jochum.dev/jo-micro/router"

	"github.com/oschwald/geoip2-golang"
	"github.com/sirupsen/logrus"
	"github.com/urfave/cli/v2"
	"jochum.dev/jo-micro/components"
)

type Config struct {
	RefreshSeconds int
	AccountID      string
	LicenseKey     string
	DataDirectory  string
}

type Handler struct {
	sync.RWMutex

	initialized bool

	cReg *components.Registry

	config Config

	refreshCancel context.CancelFunc

	dbCity    *geoip2.Reader
	dbCountry *geoip2.Reader
	dbASN     *geoip2.Reader
}

const Name = "geoipHandler"

func New() *Handler {
	return &Handler{initialized: false}
}

func MustReg(cReg *components.Registry) *Handler {
	return cReg.Must(Name).(*Handler)
}

func (h *Handler) Name() string {
	return Name
}

func (h *Handler) Priority() int {
	return 100
}

func (h *Handler) Initialized() bool {
	return h.initialized
}

func (h *Handler) Init(cReg *components.Registry, cli *cli.Context) error {
	if h.initialized {
		return nil
	}

	h.cReg = cReg

	logrus := logruscomponent.MustReg(h.cReg).Logger()

	r := router.MustReg(h.cReg)
	r.Add(
		router.NewRoute(
			router.Method(router.MethodGet),
			router.Path("/country/:ip"),
			router.Endpoint(geoippb.GeoIPService.Country),
			router.Params("ip"),
			router.AuthRequired(),
			router.RatelimitUser("1-S", "10-M"),
		),
		router.NewRoute(
			router.Method(router.MethodGet),
			router.Path("/country/:ip/:commaLanguages"),
			router.Endpoint(geoippb.GeoIPService.Country),
			router.Params("ip", "commaLanguages"),
			router.AuthRequired(),
			router.RatelimitUser("1-S", "10-M"),
		),
		router.NewRoute(
			router.Method(router.MethodGet),
			router.Path("/city/:ip"),
			router.Endpoint(geoippb.GeoIPService.City),
			router.Params("ip"),
			router.AuthRequired(),
			router.RatelimitUser("1-S", "10-M"),
		),
		router.NewRoute(
			router.Method(router.MethodGet),
			router.Path("/city/:ip/:commaLanguages"),
			router.Endpoint(geoippb.GeoIPService.City),
			router.Params("ip", "commaLanguages"),
			router.AuthRequired(),
			router.RatelimitUser("1-S", "10-M"),
		),
	)

	authVerifier := endpointroles.NewVerifier(
		endpointroles.WithLogrus(logruscomponent.MustReg(h.cReg).Logger()),
	)
	authVerifier.AddRules(
		endpointroles.NewRule(
			endpointroles.Endpoint(geoippb.GeoIPService.Country),
			endpointroles.RolesAllow(auth2.RolesServiceAndAdmin),
		),
		endpointroles.NewRule(
			endpointroles.Endpoint(geoippb.GeoIPService.City),
			endpointroles.RolesAllow(auth2.RolesServiceAndAdmin),
		),
	)
	auth2.ClientAuthMustReg(h.cReg).Plugin().AddVerifier(authVerifier)

	if cli.String(fmt.Sprintf("%s_maxmind_account_id", strings.ToLower(cReg.FlagPrefix()))) == "" || cli.String(fmt.Sprintf("%s_maxmind_license_key", strings.ToLower(cReg.FlagPrefix()))) == "" {
		err := errors.InternalServerError("NO_MAXMIND_ACCOUNT", "you have to provide geoip_maxmind_account_id and geoip_maxmind_license_key")
		logrus.Fatal(err)
		return err
	}

	if err := utils.IsDirectoryAndWriteable(cli.String(fmt.Sprintf("%s_data_directory", strings.ToLower(cReg.FlagPrefix())))); err != nil {
		logrus.Fatal(err)
		return err
	}

	h.config = Config{
		DataDirectory:  cli.String(fmt.Sprintf("%s_data_directory", strings.ToLower(cReg.FlagPrefix()))),
		RefreshSeconds: cli.Int(fmt.Sprintf("%s_refreshdb", strings.ToLower(cReg.FlagPrefix()))),
		AccountID:      cli.String(fmt.Sprintf("%s_maxmind_account_id", strings.ToLower(cReg.FlagPrefix()))),
		LicenseKey:     cli.String(fmt.Sprintf("%s_maxmind_license_key", strings.ToLower(cReg.FlagPrefix()))),
	}

	refreshContext, cancel := context.WithCancel(context.Background())
	go func(ctx context.Context) {
		ticker := time.NewTicker(time.Duration(h.config.RefreshSeconds) * time.Second)
		client, err := maxmind.New(maxmind.Config{
			Logrus:            logrus,
			DownloadDirectory: h.config.DataDirectory,
			LicenseKey:        h.config.LicenseKey,
		})
		if err != nil {
			logrus.Error(err)
			return
		}
		for {
			downloader, err := client.NewDownloader(maxmind.EIDGeoLite2Country)
			if err != nil {
				logrus.Error(err)
				return
			}
			_, err = downloader.Download()
			if err != nil {
				logrus.Error(err)
			} else {
				db, err := geoip2.Open(filepath.Join(h.config.DataDirectory, "GeoLite2-Country.mmdb"))
				if err != nil {
					logrus.Error(err)
				} else {
					h.Lock()
					h.dbCountry = db
					h.Unlock()
				}
			}

			downloader, err = client.NewDownloader(maxmind.EIDGeoLite2City)
			if err != nil {
				logrus.Error(err)
				return
			}
			_, err = downloader.Download()
			if err != nil {
				logrus.Error(err)
				continue
			} else {
				db, err := geoip2.Open(filepath.Join(h.config.DataDirectory, "GeoLite2-City.mmdb"))
				if err != nil {
					logrus.Error(err)
				} else {
					h.Lock()
					h.dbCity = db
					h.Unlock()
				}
			}

			downloader, err = client.NewDownloader(maxmind.EIDGeoLite2ASN)
			if err != nil {
				logrus.Error(err)
				return
			}
			_, err = downloader.Download()
			if err != nil {
				logrus.Error(err)
				continue
			} else {
				db, err := geoip2.Open(filepath.Join(h.config.DataDirectory, "GeoLite2-ASN.mmdb"))
				if err != nil {
					logrus.Error(err)
				} else {
					h.Lock()
					h.dbASN = db
					h.Unlock()
				}
			}

			select {
			case <-ctx.Done():
				return
			case <-ticker.C:
				continue
			}
		}
	}(refreshContext)

	h.refreshCancel = cancel

	if err := geoippb.RegisterGeoIPServiceHandler(h.cReg.Service().Server(), h); err != nil {
		return err
	}

	h.initialized = true
	return nil
}

func (h *Handler) Stop() error {
	h.refreshCancel()

	h.Lock()
	if h.dbCountry != nil {
		if err := h.dbCountry.Close(); err != nil {
			logrus.Error(err)
		}
	}
	if h.dbCity != nil {
		if err := h.dbCity.Close(); err != nil {
			logrus.Error(err)
		}
	}
	if h.dbASN != nil {
		if err := h.dbASN.Close(); err != nil {
			logrus.Error(err)
		}
	}
	h.Unlock()

	h.initialized = false
	return nil
}

func (h *Handler) Flags(r *components.Registry) []cli.Flag {
	return []cli.Flag{
		&cli.IntFlag{
			Name:    fmt.Sprintf("%s_refreshdb", strings.ToLower(r.FlagPrefix())),
			Usage:   "Refresh the DB every x seconds, default is every half day",
			EnvVars: []string{fmt.Sprintf("%s_REFRESHDB", strings.ToUpper(r.FlagPrefix()))},
			Value:   43200,
		},
		&cli.StringFlag{
			Name:    fmt.Sprintf("%s_data_directory", strings.ToLower(r.FlagPrefix())),
			Usage:   "GeoIP data directory",
			EnvVars: []string{fmt.Sprintf("%s_DATA_DIRECTORY", strings.ToUpper(r.FlagPrefix()))},
			Value:   "/data",
		},
		&cli.StringFlag{
			Name:    fmt.Sprintf("%s_maxmind_account_id", strings.ToLower(r.FlagPrefix())),
			Usage:   "Maxmind account ID",
			EnvVars: []string{fmt.Sprintf("%s_MAXMIND_ACCOUNT_ID", strings.ToUpper(r.FlagPrefix()))},
		},
		&cli.StringFlag{
			Name:    fmt.Sprintf("%s_maxmind_license_key", strings.ToLower(r.FlagPrefix())),
			Usage:   "Maxmind license key",
			EnvVars: []string{fmt.Sprintf("%s_MAXMIND_LICENSE_KEY", strings.ToUpper(r.FlagPrefix()))},
		},
	}
}

func (h *Handler) Health(context context.Context) error {
	return nil
}

func (h *Handler) WrapHandlerFunc(ctx context.Context, req server.Request, rsp interface{}) error {
	return nil
}

func (h *Handler) Country(ctx context.Context, in *geoippb.IpRequest, out *geoippb.CountryResponse) error {
	ip := net.ParseIP(in.Ip)
	if ip == nil {
		return errors.BadRequest("INVALID_IP", "invalid ip '%s' given", in.Ip)
	}

	h.RLock()
	if h.dbCountry == nil {
		h.RUnlock()
		return errors.InternalServerError("DB_NOT_READY_YET", "maxmind country database is not ready yet")
	}
	record, err := h.dbCountry.Country(ip)
	h.RUnlock()

	if err != nil {
		logrus.Error(err)
	}

	out.IsoCode = record.Country.IsoCode

	out.CountryName = make(map[string]string)
	for _, lang := range in.Languages {
		if n, ok := record.Country.Names[lang]; ok {
			out.CountryName[lang] = n
		}
	}

	if in.CommaLanguages != "" {
		for _, lang := range strings.Split(in.CommaLanguages, ",") {
			if n, ok := record.Country.Names[lang]; ok {
				out.CountryName[lang] = n
			}
		}
	}

	return nil
}

func (h *Handler) City(ctx context.Context, in *geoippb.IpRequest, out *geoippb.CityResponse) error {
	ip := net.ParseIP(in.Ip)
	if ip == nil {
		return errors.BadRequest("INVALID_IP", "invalid ip '%s' given", in.Ip)
	}

	h.RLock()
	if h.dbCity == nil {
		h.RUnlock()
		return errors.InternalServerError("DB_NOT_READY_YET", "maxmind country database is not ready yet")
	}
	record, err := h.dbCity.City(ip)
	h.RUnlock()

	if err != nil {
		logrus.Error(err)
	}

	out.IsoCode = record.Country.IsoCode
	out.TimeZone = record.Location.TimeZone
	out.Latitude = record.Location.Latitude
	out.Longitude = record.Location.Longitude

	out.CountryName = make(map[string]string)
	for _, lang := range in.Languages {
		if n, ok := record.Country.Names[lang]; ok {
			out.CountryName[lang] = n
		}
	}

	if in.CommaLanguages != "" {
		for _, lang := range strings.Split(in.CommaLanguages, ",") {
			if n, ok := record.Country.Names[lang]; ok {
				out.CountryName[lang] = n
			}
		}
	}

	out.CityName = make(map[string]string)
	for _, lang := range in.Languages {
		if n, ok := record.City.Names[lang]; ok {
			out.CityName[lang] = n
		}
	}

	if in.CommaLanguages != "" {
		for _, lang := range strings.Split(in.CommaLanguages, ",") {
			if n, ok := record.City.Names[lang]; ok {
				out.CityName[lang] = n
			}
		}
	}

	return nil
}
